from functools import wraps

import pandas as pd
from sklearn.utils.validation import check_array


def holdattr(func):
    """Return wrapper function that holds attributes of the DataFrame.
    """

    @wraps(func)
    def wrapper(self, X, **kargs):
        isdataframe         = isinstance(X, pd.DataFrame)

        if isdataframe:
            index           = X.index
            columns         = X.columns

        arr                 = func(self, X, **kargs)

        if isdataframe:
            if hasattr(self, 'shift') and hasattr(self, 'window'):
                index       = index[self.window - 1::self.shift]

            if arr.ndim == 1:
                arr         = pd.Series(
                    data    = arr,
                    index   = index
                )

            else:
                arr         = pd.DataFrame(
                    data    = arr,
                    index   = index,
                    columns = columns
                )

        return arr

    return wrapper


def window_generator(X, window=1, shift=1):
    """Generator that yields windows from given data.

    parameters
    ----------
    X : array-like, shpae = (n_samples, n_features)
        Samples.

    window : integer
        Window size.

    shift : integer
        Shift size.

    Returns
    -------
    gen : generator
        Generator.
    """

    if window < shift:
        raise ValueError('window must be greater than or equal to shift.')

    X            = check_array(X)
    n_samples, _ = X.shape

    for i in range((n_samples - window + shift) // shift):
        yield X[i * shift:i * shift + window]



import unittest

import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError

from kenchi.outlier_detection import VMFOutlierDetector


class VMFOutlierDetectorTest(unittest.TestCase):
    def setUp(self):
        n_samples  = 1000
        n_features = 10
        rnd        = np.random.RandomState(0)
        self.X     = rnd.multivariate_normal(
            mean   = np.zeros(n_features),
            cov    = np.eye(n_features),
            size   = n_samples
        )
        self.df    = pd.DataFrame(self.X)
        self.sut   = VMFOutlierDetector()

    def test_fit(self):
        self.assertIsInstance(self.sut.fit(self.X), VMFOutlierDetector)

    def test_fit_predict_ndarray(self):
        self.assertIsInstance(self.sut.fit_predict(self.X), np.ndarray)

    def test_fit_predict_dataframe(self):
        self.assertIsInstance(self.sut.fit_predict(self.df), pd.Series)

    def test_decision_function_notfitted(self):
        with self.assertRaises(NotFittedError):
            self.sut.decision_function(self.X)

    def test_predict_notfitted(self):
        with self.assertRaises(NotFittedError):
            self.sut.predict(self.X)



import numpy as np
from scipy.stats import chi2
from sklearn.base import BaseEstimator
from sklearn.preprocessing import Normalizer
from sklearn.utils.validation import check_array, check_is_fitted

from ..base import DetectorMixin
from ..utils import holdattr


class VMFOutlierDetector(BaseEstimator, DetectorMixin):
    """Outlier detector in Von Mises–Fisher distribution.

    Parameters
    ----------
    assume_normalized : bool
        If False, data are normalized before computation.

    fpr : float
        False positive rate. Used to compute the threshold.

    Attributes
    ----------
    mean_direction_ : ndarray, shape = (n_features)
        Mean direction.

    threshold_ : float
        Threshold.
    """

    def __init__(self, assume_normalized=False, fpr=0.01):
        self.assume_normalized = assume_normalized
        self.fpr               = fpr

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        self : object
            Return self.
        """

        X                    = check_array(X)

        if not self.assume_normalized:
            self._normalizer = Normalizer().fit(X)
            X                = self._normalizer.transform(X)

        mean                 = np.mean(X, axis=0)
        self.mean_direction_ = mean / np.linalg.norm(mean)

        scores               = self.decision_function(X)
        mo1                  = np.mean(scores)
        mo2                  = np.mean(scores ** 2)
        m_mo                 = 2.0 * mo1 ** 2 / (mo2 - mo1 ** 2)
        s_mo                 = 0.5 * (mo2 - mo1 ** 2) / mo1
        self.threshold_      = chi2.ppf(1.0 - self.fpr, m_mo, scale=s_mo)

        return self

    @holdattr
    def decision_function(self, X):
        """Compute the anomaly score.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        scores : array-like, shape = (n_samples)
            Anomaly score for test samples.
        """

        check_is_fitted(self, ['mean_direction_'])

        X     = check_array(X)

        if not self.assume_normalized:
            X = self._normalizer.transform(X)

        return 1.0 - X @ self.mean_direction_



import numpy as np
from scipy.stats import chi2
from sklearn.base import BaseEstimator
from sklearn.covariance import graph_lasso, MinCovDet
from sklearn.utils.validation import check_array, check_is_fitted

from ..base import DetectorMixin
from ..utils import holdattr


class GaussianOutlierDetector(BaseEstimator, DetectorMixin):
    """Outlier detector in Gaussian distribution.

    Parameters
    ----------
    assume_centered : bool
        If True, data are not centered before computation.

    fpr : float
        False positive rate. Used to compute the threshold.

    random_state : int, RandomState instance or None
        Seed of the pseudo random number generator to use when shuffling the
        data.

    support_fraction : float
        Proportion of points to be included in the support of the raw MCD
        estimate.

    use_method_of_moments : bool
        If True, the method of moments is used to compute the threshold.

    Attributes
    ----------
    threshold_ : float
        Threshold.
    """

    def __init__(
        self,                  assume_centered=False,
        fpr=0.01,              random_state=None,
        support_fraction=None, use_method_of_moments=False
    ):
        self.assume_centered       = assume_centered
        self.fpr                   = fpr
        self.random_state          = random_state
        self.support_fraction      = support_fraction
        self.use_method_of_moments = use_method_of_moments

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        self : object
            Return self.
        """

        X                    = check_array(X)
        _, n_features        = X.shape

        self._mcd            = MinCovDet(
            assume_centered  = self.assume_centered,
            random_state     = self.random_state,
            support_fraction = self.support_fraction
        ).fit(X)

        if self.use_method_of_moments:
            scores           = self._mcd.dist_
            mo1              = np.mean(scores)
            mo2              = np.mean(scores ** 2)
            m_mo             = 2.0 * mo1 ** 2 / (mo2 - mo1 ** 2)
            s_mo             = 0.5 * (mo2 - mo1 ** 2) / mo1
            self.threshold_  = chi2.ppf(1.0 - self.fpr, m_mo, scale=s_mo)

        else:
            self.threshold_  = chi2.ppf(1.0 - self.fpr, n_features, scale=1.0)

        return self

    @holdattr
    def decision_function(self, X):
        """Compute the anomaly score.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        scores : array-like, shape = (n_samples)
            Anomaly score for test samples.
        """

        check_is_fitted(self, ['_mcd'])

        X = check_array(X)

        return self._mcd.mahalanobis(X)


class GGMOutlierDetector(BaseEstimator, DetectorMixin):
    """Outlier detector using Gaussian graphical models.

    Parameters
    ----------
    alpha : float
        Regularization parameter.

    assume_centered : bool
        If True, data are not centered before computation.

    fpr : float
        False positive rate. Used to compute the threshold.

    max_iter : integer
        Maximum number of iterations.

    random_state : integer, RandomState instance or None
        Seed of the pseudo random number generator to use when shuffling the
        data.

    support_fraction : float
        Proportion of points to be included in the support of the raw MCD
        estimate.

    tol : float
        The tolerance to declare convergence. If the dual gap goes below this
        value, iterations are stopped.

    Attributes
    ----------
    covariance_ : ndarray, shape = (n_features, n_features)
        Estimated covariance matrix.

    precision_ : ndarray, shape = (n_features, n_features)
        Estimated pseudo inverse matrix.

    threshold_ : ndarray, shape = (n_features)
        Threshold.
    """

    def __init__(
        self,                  alpha=0.01,
        assume_centered=False, max_iter=100,
        fpr=0.01,              random_state=None,
        support_fraction=None, tol=0.0001
    ):
        self.alpha            = alpha
        self.assume_centered  = assume_centered
        self.max_iter         = max_iter
        self.fpr              = fpr
        self.random_state     = random_state
        self.support_fraction = support_fraction
        self.tol              = tol

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        self : object
            Return self.
        """

        X                                 = check_array(X)

        self._mcd                         = MinCovDet(
            assume_centered               = self.assume_centered,
            random_state                  = self.random_state,
            support_fraction              = self.support_fraction
        ).fit(X)

        self.covariance_, self.precision_ = graph_lasso(
            emp_cov                       = self._mcd.covariance_,
            alpha                         = self.alpha,
            max_iter                      = self.max_iter,
            tol                           = self.tol
        )

        scores                            = self.decision_function(X)
        self.threshold_                   = np.percentile(
            a                             = scores,
            q                             = 100.0 * (1.0 - self.fpr),
            axis                          = 0
        )

        return self

    @holdattr
    def decision_function(self, X):
        """Compute the anomaly score.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        scores : array-like, shape = (n_samples, n_features)
            Anomaly score for test samples.
        """

        check_is_fitted(self, ['_mcd'])

        X = check_array(X)

        return 0.5 * np.log(
            2.0 * np.pi / np.diag(self.precision_)
        ) + 0.5 / np.diag(
            self.precision_
        ) * ((X - self._mcd.location_) @ self.precision_) ** 2



import numpy as np
from sklearn.utils.validation import check_array, check_is_fitted

from .utils import holdattr


class DetectorMixin:
    """Mixin class for all detectors.
    """

    _estimator_type = 'detector'

    def fit_predict(self, X, y=None, **fit_param):
        """Fit the model to the training set X and return the labels (0
        inlier, 1 outlier) on the training set.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        is_outlier : ndarray, shape = (n_samples) or (n_windows)
            Return 0 for inliers and 1 for outliers.
        """

        return self.fit(X, y, **fit_param).predict(X)

    @holdattr
    def predict(self, X):
        """Predict if a particular sample is an outlier or not.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        is_outlier : array-like, shape = (n_samples) or (n_windows)
            Return 0 for inliers and 1 for outliers.
        """

        check_is_fitted(self, ['threshold_'])

        X = check_array(X)

        if isinstance(self.threshold_, float):
            return (
                self.decision_function(X) > self.threshold_
            ).astype(np.int32)

        else:
            return np.any(
                self.decision_function(X) > self.threshold_, axis=1
            ).astype(np.int32)



import numpy as np
from sklearn.base import BaseEstimator
from sklearn.neighbors import NearestNeighbors
from sklearn.utils.validation import check_array, check_is_fitted

from ..base import DetectorMixin
from ..utils import holdattr


class EmpiricalOutlierDetector(BaseEstimator, DetectorMixin):
    """Outlier detector using the k-nearest neighbors algorithm.

    Parameters
    ----------
    fpr : float
        False positive rate. Used to compute the threshold.

    n_jobs : integer
        Number of jobs to run in parallel. If -1, then the number of jobs is
        set to the number of CPU cores. Doesn't affect fit method.

    n_neighbors : integer
        Number of neighbors.

    p : integer
        Power parameter for the Minkowski metric.

    Attributes
    ----------
    threshold_ : float
        Threshold.
    """

    def __init__(self, fpr=0.01, n_jobs=1, n_neighbors=5, p=2):
        self.fpr         = fpr
        self.n_jobs      = n_jobs
        self.n_neighbors = n_neighbors
        self.p           = p

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        self : object
            Return self.
        """

        X               = check_array(X)

        self._neigh     = NearestNeighbors(
            metric      = 'minkowski',
            n_jobs      = self.n_jobs,
            n_neighbors = self.n_neighbors,
            p           = self.p
        ).fit(X)

        scores          = self.decision_function(X)
        self.threshold_ = np.percentile(scores, 100.0 * (1.0 - self.fpr))

        return self

    @holdattr
    def decision_function(self, X):
        """Compute the anomaly score.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        scores : array-like, shape = (n_samples)
            The anomaly score for test samples.
        """

        check_is_fitted(self, ['_neigh'])

        X             = check_array(X)
        _, n_features = X.shape

        dist, _       = self._neigh.kneighbors(X)
        radius        = np.max(dist, axis=1)

        return -np.log(self.n_neighbors) + n_features * np.log(radius)



import numpy as np
from scipy.stats import multivariate_normal
from sklearn.base import BaseEstimator
from sklearn.mixture import GaussianMixture
from sklearn.utils.validation import check_array, check_is_fitted

from ..base import DetectorMixin
from ..utils import holdattr


class GaussianMixtureOutlierDetector(BaseEstimator, DetectorMixin):
    """Outlier detector using Gaussian mixture models.

    Parameters
    ----------
    fpr : float
        False positive rate. Used to compute the threshold.

    max_iter : integer
        Maximum number of iterations.

    means_init : array-like, shape = (n_components, n_features)
        User-provided initial means.

    n_components : integer
        Number of mixture components.

    precisions_init : array-like
        User-provided initial precisions.

    random_state : integer, RandomState instance or None
        Seed of the pseudo random number generator to use when shuffling the
        data.

    tol : float
        Convergence threshold.

    weights_init : array-like, shape = (n_components)
        User-provided initial weights.

    Attributes
    ----------
    weights_ : ndarray, shape = (n_components)
        Weight of each mixture component.

    means_ : ndarray, shape = (n_components, n_features)
        Mean of each mixture component.

    covariances_ : ndarray
        Covariance of each mixture component.

    precisions_ : ndarray
        Precision matrix of each mixture component.

    threshold_ : float
        Threshold.
    """

    def __init__(
        self,              fpr=0.01,
        max_iter=100,      means_init=None,
        n_components=1,    precisions_init=None,
        random_state=None, tol=1e-03,
        weights_init=None
    ):
        self.fpr             = fpr
        self.max_iter        = max_iter
        self.means_init      = means_init
        self.n_components    = n_components
        self.precisions_init = precisions_init
        self.random_state    = random_state
        self.tol             = tol
        self.weights_init    = weights_init

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        self : object
            Return self.
        """

        X                   = check_array(X)

        gmm                 = GaussianMixture(
            max_iter        = self.max_iter,
            means_init      = self.means_init,
            n_components    = self.n_components,
            precisions_init = self.precisions_init,
            random_state    = self.random_state,
            tol             = self.tol,
            weights_init    = self.weights_init
        ).fit(X)

        self.weights_       = gmm.weights_
        self.means_         = gmm.means_
        self.covariances_   = gmm.covariances_
        self.precisions_    = gmm.precisions_

        scores              = self.decision_function(X)
        self.threshold_     = np.percentile(scores, 100.0 * (1.0 - self.fpr))

        return self

    @holdattr
    def decision_function(self, X):
        """Compute the anomaly score.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        scores : array-like, shape = (n_samples)
            Anomaly score for test samples.
        """

        check_is_fitted(
            self, ['weights_', 'means_', 'covariances_', 'precisions_']
        )

        X = check_array(X)

        return -np.log(
            np.sum([
                weight * multivariate_normal.pdf(
                    X, mean=mean, cov=cov
                ) for weight, mean, cov in zip(
                    self.weights_, self.means_, self.covariances_
                )
            ], axis=0)
        )



import numpy as np
from sklearn.base import BaseEstimator
from sklearn.covariance import graph_lasso, MinCovDet
from sklearn.utils.validation import check_array, check_is_fitted

from ..base import DetectorMixin
from ..utils import holdattr, window_generator


class GGMChangeDetector(BaseEstimator, DetectorMixin):
    """Change detector using Gaussian graphical models.

    Parameters
    ----------
    alpha : float
        Regularization parameter.

    assume_centered : bool
        If True, data are not centered before computation.

    fpr : float
        False positive rate. Used to compute the threshold.

    max_iter : integer
        Maximum number of iterations.

    random_state : integer, RandomState instance or None
        Seed of the pseudo random number generator to use when shuffling the
        data.

    shift : integer
        Shift size.

    support_fraction : float
        Proportion of points to be included in the support of the raw MCD
        estimate.

    tol : float
        The tolerance to declare convergence. If the dual gap goes below this
        value, iterations are stopped.

    window : integer
        Window size.

    Attributes
    ----------
    covariance_ : ndarray, shape = (n_features, n_features)
        Estimated covariance matrix.

    precision_ : ndarray, shape = (n_features, n_features)
        Estimated pseudo inverse matrix.

    threshold_ : ndarray, shape = (n_features)
        Threshold.
    """

    def __init__(
        self,                  alpha=0.01,
        assume_centered=False, max_iter=100,
        fpr=0.01,              random_state=None,
        shift=50,              support_fraction=None,
        tol=0.0001,            window=100
    ):
        self.alpha            = alpha
        self.assume_centered  = assume_centered
        self.max_iter         = max_iter
        self.fpr              = fpr
        self.random_state     = random_state
        self.shift            = shift
        self.support_fraction = support_fraction
        self.tol              = tol
        self.window           = window

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Samples.

        Returns
        -------
        self : object
            Return self.
        """

        X                                 = check_array(X)

        self._mcd                         = MinCovDet(
            assume_centered               = self.assume_centered,
            random_state                  = self.random_state,
            support_fraction              = self.support_fraction
        ).fit(X)

        self.covariance_, self.precision_ = graph_lasso(
            emp_cov                       = self._mcd.covariance_,
            alpha                         = self.alpha,
            max_iter                      = self.max_iter,
            tol                           = self.tol
        )

        scores                            = self.decision_function(X)
        self.threshold_                   = np.percentile(
            a                             = scores,
            q                             = 100.0 * (1.0 - self.fpr),
            axis                          = 0
        )

        return self

    @holdattr
    def decision_function(self, X):
        """Compute the anomaly score.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        scores : array-like, shape = (n_windows, n_features)
            Anomaly score for test samples.
        """

        check_is_fitted(self, ['_mcd'])

        X                        = check_array(X)
        n_samples, n_features    = X.shape

        scores                   = np.empty(
            ((n_samples - self.window + self.shift) // self.shift, n_features)
        )

        for i, X_window in enumerate(
            window_generator(X, self.window, self.shift)
        ):
            mcd_window           = MinCovDet(
                assume_centered  = self.assume_centered,
                random_state     = self.random_state,
                support_fraction = self.support_fraction
            ).fit(X_window)

            _, precision_window  = graph_lasso(
                emp_cov          = mcd_window.covariance_,
                alpha            = self.alpha,
                max_iter         = self.max_iter,
                tol              = self.tol
            )

            scores[i]            = 0.5 * np.log(
                np.diag(self.precision_) / np.diag(precision_window)
            ) - 0.5 * (
                np.diag(
                    self.precision_ @ self._mcd.covariance_ @ self.precision_
                ) / np.diag(self.precision_) - np.diag(
                    precision_window @ self._mcd.covariance_ @ precision_window
                ) / np.diag(precision_window)
            )

        return scores



from .gaussian_distns import GaussianOutlierDetector, GGMOutlierDetector
from .empirical_distns import EmpiricalOutlierDetector
from .mixture_distns import GaussianMixtureOutlierDetector
from .vmf_distns import VMFOutlierDetector



from setuptools import find_packages, setup

from kenchi import __version__


with open('README.rst') as f:
    readme           = f.read()

with open('requirements.txt') as f:
    requires         = f.read().splitlines()

setup(
    name             = 'kenchi',
    version          = __version__,
    author           = 'Kon',
    author_email     = 'kon.y.ohr.n@gmail.com',
    url              = 'http://kenchi.readthedocs.io',
    description      = 'A set of python modules for anomaly detection',
    long_description = readme,
    license          = 'MIT',
    packages         = find_packages(exclude=['tests']),
    install_requires = requires,
    test_suite       = 'kenchi.tests.suite'
)
